import unittest

from xpdt import NameSpace, StructDecl, MemberDecl, BaseType

class Test_Nested(unittest.TestCase):
    def setUp(self):
        self.code = NameSpace.from_decls(
            (
                StructDecl(
                    struct_name='Point',
                    members=[
                        MemberDecl('x', 's32'),
                        MemberDecl('y', 's32'),
                        MemberDecl('z', 's32'),
                    ],
                ),
                StructDecl(
                    struct_name='Entity',
                    members=[
                        MemberDecl('id', 'u32'),
                        MemberDecl('pos', 'Point'),
                    ],
                ),
            ),
            name=None,
        ).gen_dynamic_python()

    def test_point(self):
        point = self.code.Point(1, 2, 3)
        orig = self.code.Entity(0xdeadbeef, point)
        b = bytes(orig)
        clone = orig._frombytes(b)
        self.assertEqual(orig, clone)



from typing import Dict
from enum import Enum

from .integraltype import IntegralType
from .buftype import BufType
from .uuidtype import UuidType
from .typedef import TypeDef

__all__ = (
    'BaseType',
    'base_types',
)

Blob = BufType()
xu128 = UuidType()
UnsignedInt8 = IntegralType(8, False)
SignedInt8 = IntegralType(8, True)
UnsignedInt16 = IntegralType(16, False)
SignedInt16 = IntegralType(16, True)
UnsignedInt32 = IntegralType(32, False)
SignedInt32 = IntegralType(32, True)
UnsignedInt64 = IntegralType(64, False)
SignedInt64 = IntegralType(64, True)


class BaseType(TypeDef, Enum):
    __slots__ = ()

    blob = 'blob', Blob
    u8 = 'u8', UnsignedInt8
    s8 = 's8', SignedInt8
    u16 = 'u16', UnsignedInt16
    s16 = 's16', SignedInt16
    u32 = 'u32', UnsignedInt32
    s32 = 's32', SignedInt32
    u64 = 'u64', UnsignedInt64
    s64 = 's64', SignedInt64
    u128 = 'u128', xu128
    uuid = 'uuid', xu128


def base_types() -> Dict[str, TypeDef]:
    return {t.name: t.value for t in BaseType}



from typing import NamedTuple, Generator, FrozenSet, Pattern
import re


__all__ = (
    'lex',
    'Lexeme',
)


_re = re.compile('|'.join((
    r'(?P<kw_type>type)',
    r'(?P<identifier>[a-zA-Z_][a-zA-Z0-9_]*)',
    r'(?P<num>[0-9][0-9]*)',
    r'(?P<lbrace>\{)',
    r'(?P<rbrace>\})',
    r'(?P<semicolon>;)',
    r'(?P<comment>#.*?(?=\n))',          # up to but not including newline
    r'(?P<whitespace>[^\S\n][^\S\n]*)',  # non-newline whitespace
    r'(?P<newline>\n)',
)))
_ignore_toks = frozenset({'whitespace', 'comment', 'newline'})


class Lexeme(NamedTuple):
    tok_type: str
    val: str
    file: str
    line: int


def lex(s: str,
        file: str = '<string>',
        line: int = 1,
        eof: bool = True,
        _ignored: FrozenSet[str] = _ignore_toks,
        _dfa: Pattern[str] = _re,
        ) -> Generator[Lexeme, None, None]:
    pos = 0

    while True:
        m = _dfa.match(s, pos)
        if not m:
            break
        grp = m.lastgroup
        begin, end = m.span()
        pos = end
        if grp not in _ignored:
            assert(grp is not None)
            yield Lexeme(grp, s[begin:end], file, line)
        elif grp == 'newline':
            line += 1

    if pos < len(s):
        raise ValueError

    if eof:
        yield Lexeme('eof', '$', file, line)



from typing import (
    Optional, Tuple, Dict, NamedTuple, Type, TextIO, Iterable, Iterator, Any
)
from types import SimpleNamespace

from .dupes import dupes
from .decl import StructDecl, MemberDecl
from .typedef import TypeDef
from .basetypes import base_types
from .member import MemberDef
from .struct import StructDef, StructType
from .jinja import ctemplate, pytemplate
from .c import SysInclude

__all__ = (
    'NameSpace',
)


class NameSpace:
    __slots__ = (
        '_name',
        '_structs',
    )

    _name: Optional[str]
    _structs: Dict[str, StructDef]

    x1b_c = ctemplate.get_template('x1b.c')
    x1b_python = pytemplate.get_template('x1b.pyt')

    def __init__(self,
                 structs: Tuple[StructDef, ...],
                 name: Optional[str] = None):

        d = dupes((s.name for s in structs))
        if d:
            raise ValueError(f'{name}: Duplicated struct name: '
                             ", ".join(d))

        self._name = name

        # TODO: Allow sub-structs
        self._structs = {s.name: s for s in structs}

    @property
    def name(self) -> Optional[str]:
        return self._name

    def __iter__(self) -> Iterator[StructDef]:
        yield from self._structs.values()

    def __len__(self) -> int:
        return len(self._structs)

    def _gen_dynamic_clsname(self, s: StructDef) -> str:
        if self._name:
            return f'{self._name}.{s.name}'
        else:
            return s.name

    def _gen_storage_class(self, s: StructDef) -> Type[NamedTuple]:
        sig = [(m.name, m.typedef.type.pytype) for m in s]

        # This class is a tuple for storing the data in each instance
        # mypy insists on literals, but sorry, no can do...
        return NamedTuple(s.name, sig)  # type: ignore

    @property
    def python_code(self) -> str:
        return self.x1b_python.render(namespace=self)

    def gen_dynamic_python(self) -> SimpleNamespace:
        names = {s.name for s in self}
        g: Dict[str, Any] = {}
        exec(self.python_code, g, g)
        return SimpleNamespace(**{x: g[x] for x in names})

    def gen_python(self, f: TextIO) -> None:
        f.write(self.python_code)

    def gen_c(self, f: TextIO) -> None:
        x1b_hdr = SysInclude('xpdt/x1b.h')
        headers = (
            SysInclude('stdint.h'),
            SysInclude('stdbool.h'),
            SysInclude('string.h'),
        )

        code = self.x1b_c.render(
            headers=headers,
            namespace=self,
            x1b_hdr=x1b_hdr,
        )
        f.write(code)

    @classmethod
    def from_decls(cls,
                   decls: Iterable[StructDecl],
                   name: Optional[str] = None) -> 'NameSpace':

        typemap = base_types()

        def define_member(decl: MemberDecl) -> MemberDef:
            type_name = decl.type_name
            try:
                member_type = typemap[type_name]
            except KeyError:
                raise KeyError(f'Type "{type_name}" not known')
            return MemberDef(
                decl.member_name,
                member_type,
            )

        def define_struct(decl: StructDecl) -> StructDef:
            name = decl.struct_name
            struct = StructDef(
                name,
                tuple((define_member(m) for m in decl.members)),
            )
            if name in typemap:
                raise ValueError(f'Type {name} already defined')
            typemap[name] = TypeDef(name, StructType(struct))
            return struct

        defns = tuple((define_struct(decl) for decl in decls))

        return cls(defns, name)



from typing import NamedTuple, List

__all__ = (
    'MemberDecl',
    'StructDecl',
)


class MemberDecl(NamedTuple):
    member_name: str
    type_name: str


class StructDecl(NamedTuple):
    struct_name: str
    members: List[MemberDecl]



from .type import XpdtType

__all__ = (
    'IntegralType',
)


class IntegralType(XpdtType):
    __slots__ = (
        '_width',
        '_signed',
        '_fmt',
        '_size',
        '_ctype',
    )

    pytype = 'int'

    _struct_map = {
        8: 'b',
        16: 'h',
        32: 'i',
        64: 'q',
    }

    def __init__(self, width: int, signed: bool):
        self._width = width
        self._signed = signed
        fmt = self._struct_map[self._width]
        if signed:
            self._ctype = f'int{width}_t'
            fmt = fmt.lower()
        else:
            self._ctype = f'uint{width}_t'
            fmt = fmt.upper()
        self._fmt = fmt
        self._size = super().size

    @property
    def ctype(self) -> str:
        return self._ctype

    @property
    def struct_fmt(self) -> str:
        return self._fmt

    @property
    def size(self) -> int:
        return self._size

    @property
    def needs_vbuf(self) -> bool:
        return False



from __future__ import annotations

from typing import NamedTuple, Generator, Tuple, TYPE_CHECKING
from enum import Enum
from abc import ABC, abstractmethod
from struct import calcsize


if TYPE_CHECKING:
    from .member import MemberDef


__all__ = (
    'ConstructAction',
    'ConstructElement',
    'XpdtType',
)


class ConstructAction(Enum):
    AGG_PUSH = 'agg_push'
    AGG_POP = 'agg_pop'
    MEMBER = 'member'


class ConstructElement(NamedTuple):
    action: ConstructAction
    member: MemberDef
    path: Tuple[MemberDef, ...] = ()

    def indent(self, pat: str = '\t', base: int = 0) -> str:
        return pat * (base + len(self.path))

    @property
    def path_names(self) -> Tuple[str, ...]:
        return tuple((x.name for x in self.path))

    @property
    def full_path(self) -> Tuple[MemberDef, ...]:
        return self.path + (self.member, )

    @property
    def full_path_names(self) -> Tuple[str, ...]:
        return tuple((x.name for x in self.full_path))

    @property
    def c_named_initializer(self) -> str:
        return '.'.join(self.full_path_names)

    @property
    def python_var_name(self) -> str:
        # FIXME: this is not prefix-free
        return f'_{"_".join(self.full_path_names)}'


class XpdtType(ABC):
    @property
    @abstractmethod
    def struct_fmt(self) -> str:
        pass

    @property
    @abstractmethod
    def ctype(self) -> str:
        pass

    @property
    @abstractmethod
    def pytype(self) -> str:
        pass

    @property
    @abstractmethod
    def needs_vbuf(self) -> bool:
        pass

    @property
    def size(self) -> int:
        return calcsize(self.struct_fmt)

    @property
    def is_scalar(self) -> bool:
        return True

    @property
    def struct_name(self) -> str:
        # Only relevant for aggregate types
        raise NotImplementedError

    def construct_recursive(self, path: Tuple[MemberDef, ...] = ()) \
            -> Generator[ConstructElement, None, None]:
        # Only relevant for aggregate types
        raise NotImplementedError

    @property
    def scalar_members(self) -> Generator[Tuple[str, ...], None, None]:
        raise NotImplementedError



import unittest

from xpdt import NameSpace, StructDecl, MemberDecl, BaseType

class Test_Stringy(unittest.TestCase):
    def setUp(self):
        self.code = NameSpace.from_decls(
            (
                StructDecl(
                    struct_name='Item',
                    members=[
                        MemberDecl('id', 'u32'),
                        MemberDecl('name', 'blob'),
                    ],
                ),
                StructDecl(
                    struct_name='Entity',
                    members=[
                        MemberDecl('id', 'u32'),
                        MemberDecl('a', 'Item'),
                        MemberDecl('b', 'Item'),
                    ],
                ),
            ),
            name=None,
        ).gen_dynamic_python()

    def test_entity(self):
        obj_a = self.code.Item(1, b'Item A')
        obj_b = self.code.Item(2, b'Item B')
        orig = self.code.Entity(0xdeadbeef, obj_a, obj_b)
        b = bytes(orig)
        clone = orig._frombytes(b)
        self.assertEqual(orig, clone)

    def test_entity_list(self):
        obj_a = self.code.Item(1, b'Item A')
        obj_b = self.code.Item(2, b'Item B')
        obj_c = self.code.Item(3, b'Item C')
        ent_a = self.code.Entity(0xdeadbeef, obj_a, obj_b)
        ent_b = self.code.Entity(0xfeedface, obj_c, obj_b)

        orig = [ent_a, ent_b]

        clone = b''.join((bytes(ent) for ent in orig))

        tot_len = len(clone)
        decoded = []
        off = 0
        while off < tot_len:
            item_len, ent = self.code.Entity._frombuf(clone, off)
            off += item_len
            decoded.append(ent)

        self.assertListEqual(orig, decoded)



from argparse import ArgumentParser
from pathlib import Path
from sys import stdout
from itertools import chain
import logging

from xpdt import NameSpace
from .parse import parse_file
from .shiftreduce import ParseError

_fmt = logging.Formatter('%(message)s')
_stdio_handler = logging.StreamHandler(stream=stdout)
_stdio_handler.setFormatter(_fmt)
_log = logging.getLogger('xpdt')
_log.addHandler(_stdio_handler)


def gen_c(ns: NameSpace, base: Path, module_name: str) -> None:
    p = base / f'{module_name}.h'
    _log.info('Writing code: %s', p)
    with p.open('w') as f:
        ns.gen_c(f)


def gen_py(ns: NameSpace, base: Path, module_name: str) -> None:
    p = base / f'{module_name}.py'
    _log.info('Writing code: %s', p)
    with p.open('w') as f:
        ns.gen_python(f)


def main() -> None:
    opts = ArgumentParser(description='xpdt: eXPedited Data Transfer')
    opts.add_argument('--verbose', '-v',
                      action='count',
                      default=0,
                      help='Be more talkative')
    opts.add_argument('--module-name', '-n',
                      default=None,
                      type=str,
                      help='Name of module (output file)')
    opts.add_argument('--language', '-l',
                      default='c',
                      type=str,
                      help='Output language')
    opts.add_argument('--out', '-o',
                      default=Path(),
                      type=Path,
                      help='Output dir')
    opts.add_argument('file',
                      type=Path,
                      nargs='+',
                      help='xpdt schema files')

    args = opts.parse_args()
    if args.verbose:
        _log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)

    if args.module_name is None:
        if len(args.file) == 1:
            in_path, = args.file
            module_name = in_path.stem
        else:
            module_name = 'xpdt'
    else:
        module_name = args.module_name

    try:
        decls = tuple(chain(*map(parse_file, args.file)))
    except ParseError as e:
        print(e)
        raise SystemExit(1)

    ns = NameSpace.from_decls(decls, name=None)

    backends = {
        'c': gen_c,
        'py': gen_py,
        'python': gen_py,
    }

    try:
        fn = backends[args.language]
    except KeyError:
        print(f'Unknown language: "{args.language}"')
        raise SystemExit(1)

    args.out.mkdir(exist_ok=True)
    fn(ns, args.out, module_name)


if __name__ == '__main__':
    main()



import unittest

from xpdt import NameSpace, StructDef, MemberDef, BaseType

class Test_RoundTrips(unittest.TestCase):
    def setUp(self):
        self.code = NameSpace(
            structs=(
                StructDef(
                    name='Point',
                    members=(
                        MemberDef('x', BaseType.s32),
                        MemberDef('y', BaseType.s32),
                        MemberDef('z', BaseType.s32),
                    ),
                ),
                StructDef(
                    name='Item',
                    members=(
                        MemberDef('id', BaseType.u32),
                        MemberDef('first_name', BaseType.blob),
                        MemberDef('surname', BaseType.blob),
                    ),
                ),
            ),
            name=None,
        ).gen_dynamic_python()

    def test_point(self):
        orig = self.code.Point(1, -2, 3)
        b = bytes(orig)
        clone = orig._frombytes(b)
        self.assertEqual(orig, clone)


    def test_item(self):
        orig = self.code.Item(0xdeadbeef, b'Malkovich', b'')
        b = bytes(orig)
        clone = orig._frombytes(b)
        self.assertEqual(orig, clone)



from __future__ import annotations
from typing import NamedTuple, Dict, Generator, List, Optional, Union
from pathlib import Path

from .decl import StructDecl, MemberDecl
from .shiftreduce import state, State, ParseError
from .lex import lex, Lexeme

__all__ = (
    'parse',
    'parse_file',
)


class MemberTypeItem(NamedTuple):
    member_type: str


class MemberNameItem(NamedTuple):
    member_name: str


class NamespaceItem(NamedTuple):
    ns: Dict[str, StructDecl]


Item = Union[
    MemberTypeItem,
    MemberNameItem,
    NamespaceItem,
    StructDecl,
]


class Parser:
    __slots__ = (
        '_tokens',
        '_stack',
    )

    _stack: List[Item]

    def __init__(self, toks: Generator[Lexeme, None, None]):
        self._tokens = toks
        self._stack = list()

    def push(self, item: Item) -> None:
        self._stack.append(item)

    def pop(self) -> Item:
        return self._stack.pop()

    def stacktop(self) -> Item:
        return self._stack[-1]

    @state
    def _initial(self) -> None:
        pass

    @state
    def _type_name(self) -> None:
        pass

    @state
    def _type_start(self) -> None:
        pass

    @state
    def _member_type(self) -> None:
        pass

    @state
    def _member_name(self) -> None:
        pass

    @state
    def _member_end(self) -> None:
        pass

    @_initial.on('semicolon')
    def _initial_semi(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce
        return self._initial

    @_initial.on('eof')
    def _initial_eof(self, tok: Lexeme) -> Optional[State[Parser]]:
        # accept
        return None

    @_initial.on('kw_type')
    def _initial_type_keyword(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        return self._type_name

    @_type_start.on('lbrace')
    def _type_start_lbrace(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        return self._member_type

    @_type_name.on('identifier')
    def _type_name_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(StructDecl(tok.val, list()))
        return self._type_start

    @_member_type.on('identifier')
    def _member_type_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(MemberTypeItem(tok.val))
        return self._member_name

    @_member_name.on('identifier')
    def _member_name_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(MemberNameItem(tok.val))
        return self._member_end

    @_member_end.on('semicolon')
    def _member_end_semicolon(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce MemberDecl
        name_item = self.pop()
        type_item = self.pop()

        assert isinstance(name_item, MemberNameItem)
        assert isinstance(type_item, MemberTypeItem)

        member_name = name_item.member_name
        member_type = type_item.member_type

        t = self.stacktop()
        assert isinstance(t, StructDecl)

        t.members.append(MemberDecl(member_name, member_type))
        return self._member_type

    @_member_type.on('rbrace')
    def _member_type_rbrace(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce StructDef
        t = self.pop()
        ns = self.stacktop()

        assert isinstance(t, StructDecl)
        assert isinstance(ns, NamespaceItem)

        ns.ns[t.struct_name] = t
        return self._initial

    def __iter__(self) -> Generator[StructDecl, None, None]:
        self._stack.clear()
        self.push(NamespaceItem({}))

        gen = self._tokens
        st: Optional[State[Parser]] = self._initial

        while st is not None:
            try:
                tok = next(gen)
            except StopIteration:
                raise ParseError('Unexpected end of token-stream')

            st = st(tok)

        ns_item = self.pop()
        assert isinstance(ns_item, NamespaceItem)

        yield from ns_item.ns.values()

    @classmethod
    def fromstring(cls, s: str, file: str = '<string>') -> Parser:
        return cls(lex(s, file=file))

    @classmethod
    def fromfile(cls, p: Path) -> Parser:
        """
        Loads the file, line at a time.
        """

        def gen() -> Generator[Lexeme, None, None]:
            file = str(p)
            with p.open() as f:
                lno = 1
                for line in f:
                    yield from lex(line, file, lno, False)
                    lno += 1
            yield Lexeme('eof', '$', file, lno)
        return cls(gen())


def parse(s: str) -> Generator[StructDecl, None, None]:
    yield from Parser.fromstring(s)


def parse_file(p: Path) -> Generator[StructDecl, None, None]:
    yield from Parser.fromfile(p)



from timeit import default_timer as timer
from pathlib import Path

from xpdt import NameSpace, StructDef, MemberDef, BaseType


def main():
    cls = NameSpace(
        structs=(
            StructDef(
                name='SomeStruct',
                members=(
                    MemberDef('a', BaseType.u64),
                    MemberDef('b', BaseType.u64),
                    MemberDef('c', BaseType.u64),
                    MemberDef('d', BaseType.u64),
                    MemberDef('e', BaseType.u64),
                    MemberDef('f', BaseType.u64),
                    MemberDef('g', BaseType.u64),
                    MemberDef('h', BaseType.u64),
                    MemberDef('i', BaseType.u64),
                    MemberDef('j', BaseType.u64),
                    MemberDef('k', BaseType.u64),
                    MemberDef('l', BaseType.u64),
                    MemberDef('m', BaseType.u64),
                    MemberDef('n', BaseType.u64),
                    MemberDef('o', BaseType.u64),
                    MemberDef('p', BaseType.u64),
                ),
            ),
        ),
    ).gen_dynamic_python().SomeStruct

    sz = cls._bin_size

    with Path('/dev/zero').open('rb') as f:
        iters = 1000000
        start = timer()
        for i in range(iters):
            buf = f.read(sz)
            obj = cls._frombytes(buf)
        end = timer()
        elapsed = end - start
        per_iter = elapsed / iters
        print(f'{elapsed:.2f} sec elapsed')
        print(f'{per_iter*1e9:.0f} nsec per record')


if __name__ == '__main__':
    main()



from __future__ import annotations
from typing import Dict, Optional, Callable, Any, TypeVar, Generic, cast
from typing_extensions import Protocol
from functools import wraps

"""
Helpers for shift/reduce parsing
"""

__all__ = (
    'ParseError',
    'state',
)


class Token(Protocol):
    @property
    def tok_type(self) -> Any: ...

    @property
    def val(self) -> str: ...

    @property
    def file(self) -> str: ...

    @property
    def line(self) -> int: ...


class ParseError(Exception):
    tok: Optional[Token]

    def __init__(self,
                 msg: str,
                 /,
                 tok: Optional[Token] = None):
        super().__init__(msg)
        self.tok = tok

    def __str__(self) -> str:
        s = super().__str__()
        tok = self.tok
        if tok is None:
            return s
        return f'{tok.file}:{tok.line}: at "{tok.val}": {s}'


_T = TypeVar('_T')


class State(Generic[_T]):
    def on(self: _T,
           *args: str,
           ) -> Callable[..., Callable[[_T, Token], Optional[State[_T]]]]: ...

    def __call__(self: _T, tok: Token) -> Optional[State[_T]]: ...


def state(func: Callable[[_T], None]) -> State[_T]:
    registry: Dict[Any, Callable[[_T, Token], Optional[State[_T]]]] = {}

    def on(*args: str,
           ) -> Callable[..., Callable[[_T, Token], Optional[State[_T]]]]:
        def decorator(func: Callable[[_T, Token], Optional[State[_T]]],
                      ) -> Callable[[_T, Token], Optional[State[_T]]]:
            for tok_type in args:
                registry[tok_type] = func
            return func
        return decorator

    @wraps(func)
    def wrapper(self: _T, tok: Token) -> Optional[State[_T]]:
        tok_type = tok.tok_type
        try:
            f = registry[tok_type]
        except KeyError:
            expected = ', '.join(sorted(registry.keys()))
            raise ParseError(
                f'{tok_type} unexpected, looking for {expected}',
                tok=tok,
            )
        return f(self, tok)

    wrapper.on = on  # type: ignore
    return cast(State[_T], wrapper)



from jinja2 import Environment, PackageLoader, StrictUndefined
import builtins

__all__ = (
    'ctemplate',
    'pytemplate',
)


def _prefix(s: str, pfx: str) -> str:
    return pfx + s


def _suffix(s: str, sfx: str) -> str:
    return s + sfx


def _wrap(s: str, pfx: str, sfx: str) -> str:
    return pfx + s + sfx


def _macro(s: str) -> str:
    return s.replace('\n', ' \\\n')


def _setup(env: Environment) -> None:
    env.undefined = StrictUndefined
    env.filters.update({f: getattr(builtins, f) for f in (
        'len',
        'range',
        'zip',
        'hex',
        'enumerate',
        'sorted',
    )})
    env.filters.update({
        'prefix': _prefix,
        'suffix': _suffix,
        'wrap': _wrap,
        'macro': _macro,
    })


ctemplate = Environment(
    loader=PackageLoader('xpdt', 'templates'),
    line_statement_prefix='//',
    variable_start_string='/*{',
    variable_end_string='}*/',
    block_start_string='/*#',
    block_end_string='#*/',
    keep_trailing_newline=False,
    trim_blocks=True,
    lstrip_blocks=True,
    auto_reload=False,
    autoescape=False,
)
_setup(ctemplate)


pytemplate = Environment(
    loader=PackageLoader('xpdt', 'templates'),
    block_start_string='#%',
    block_end_string='%#',
    line_statement_prefix='##',
    variable_start_string='$$',
    variable_end_string='$$',
    keep_trailing_newline=False,
    trim_blocks=True,
    lstrip_blocks=True,
    auto_reload=False,
    autoescape=False,
)
_setup(pytemplate)



from .type import XpdtType
from .integraltype import IntegralType

__all__ = (
    'BufType',
)


class BufType(XpdtType):
    __slots__ = (
        '_off_type',
    )

    pytype = 'bytes'

    def __init__(self) -> None:
        self._off_type = IntegralType(32, False)

    @property
    def struct_fmt(self) -> str:
        return self._off_type.struct_fmt

    @property
    def ctype(self) -> str:
        return 'struct xbuf'

    @property
    def size(self) -> int:
        return self._off_type.size

    @property
    def needs_vbuf(self) -> bool:
        return True
